@@ -3,30 +3,48 @@
 # Copyright (C) 2011 Lincoln de Sousa <lincoln@comum.org>
 # Copyright 2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
 #
-# This program is free software: you can redistribute it and/or modify
-# it under the terms of the GNU Affero General Public License as published by
-# the Free Software Foundation, either version 3 of the License, or
-# (at your option) any later version.
+# This file is part of Flask-Restless.
 #
-# This program is distributed in the hope that it will be useful,
-# but WITHOUT ANY WARRANTY; without even the implied warranty of
-# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-# GNU Affero General Public License for more details.
+# Flask-Restless is free software: you can redistribute it and/or modify it
+# under the terms of the GNU Affero General Public License as published by the
+# Free Software Foundation, either version 3 of the License, or (at your
+# option) any later version.
+#
+# Flask-Restless is distributed in the hope that it will be useful, but WITHOUT
+# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+# FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+# details.
 #
 # You should have received a copy of the GNU Affero General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
-"""Provides querying, searching, and function evaluation on Elixir models.
+# along with Flask-Restless. If not, see <http://www.gnu.org/licenses/>.
+"""
+    flask.ext.restless.search
+    ~~~~~~~~~~~~~~~~~~~~~~~~~
 
-:copyright:2011 by Lincoln de Sousa <lincoln@comum.org>
-:copyright:2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
-:license: GNU AGPLv3, see COPYING for more details
+    Provides querying, searching, and function evaluation on Elixir models.
+
+    The most important functions in this module are the :func:`create_query`
+    and :func:`search` functions, which create a SQLAlchemy query object and
+    execute that query on a given model, respectively.
+
+    :copyright:2011 by Lincoln de Sousa <lincoln@comum.org>
+    :copyright:2012 Jeffrey Finkelstein <jeffrey.finkelstein@gmail.com>
+    :license: GNU AGPLv3, see COPYING for more details
 
 """
 
-from elixir import session
-from sqlalchemy.sql import func
-
-
+#: The mapping from operator name (as accepted by the search method) to a
+#: function which returns the SQLAlchemy expression corresponding to that
+#: operator.
+#:
+#: The function in each of the values takes three arguments. The first argument
+#: is the field object on which to apply the operator. The second argument is
+#: the second argument to the operator, should one exist. The third argument is
+#: the name of the field. All functions use the first argument, some use the
+#:  second, and few use the third.
+#:
+#: Some operations have multiple names. For example, the equality operation can
+#: be described by the strings ``'=='``, ``'eq'``, ``'equals'``, etc.
 OPERATORS = {
     '==': lambda f, a, fn: f == a,
     'eq': lambda f, a, fn: f == a,
@@ -56,23 +74,18 @@ OPERATORS = {
     'is_not_null': lambda f, a, fn: f != None,
     'desc': lambda f, a, fn: f.desc,
     'asc': lambda f, a, fn: f.asc,
-    'has': lambda f, a, fn: f.has(**{fn: a}),
-    'any': lambda f, a, fn: f.any(**{fn: a})
+    # HACK For Python 2.5, unicode dictionary keys are not allowed.
+    'has': lambda f, a, fn: f.has(**{str(fn): a}),
+    'any': lambda f, a, fn: f.any(**{str(fn): a})
 }
-"""The mapping from operator name (as accepted by the search method) to a
-function which returns the SQLAlchemy expression corresponding to that
-operator.
 
-The function in each of the values takes three arguments. The first argument is
-the field object on which to apply the operator. The second argument is the
-second argument to the operator, should one exist. The third argument is the
-name of the field. All functions use the first argument, some use the second,
-and few use the third.
 
-Some operations have multiple names. For example, the equality operation can be
-described by the strings ``'=='``, ``'eq'``, ``'equals'``, etc.
+def _unicode_keys_to_strings(dictionary):
+    """Returns a new dictionary with the same mappings as `dictionary`, but
+    with each of the keys coerced to a string (by calling :func:`str(key)`).
 
-"""
+    """
+    return dict((str(k), v) for k, v in dictionary.iteritems())
 
 
 class IllegalArgumentError(Exception):
@@ -179,7 +192,7 @@ class SearchParameters(object):
 
     """
 
-    def __init__(self, filters=[], limit=None, offset=None, order_by=[]):
+    def __init__(self, filters=None, limit=None, offset=None, order_by=None):
         """Instantiates this object with the specified attributes.
 
         `filters` is a list of :class:`Filter` objects, representing filters to
@@ -196,10 +209,10 @@ class SearchParameters(object):
         search.
 
         """
-        self.filters = filters
+        self.filters = filters or []
         self.limit = limit
         self.offset = offset
-        self.order_by = order_by
+        self.order_by = order_by or []
 
     def __repr__(self):
         """Returns a string representation of the search parameters."""
@@ -239,7 +252,10 @@ class SearchParameters(object):
         from_dict = Filter.from_dictionary
         # may raise IllegalArgumentError
         filters = [from_dict(f) for f in dictionary.get('filters', [])]
-        order_by = [OrderBy(**o) for o in dictionary.get('order_by', [])]
+        # HACK In Python 2.5, unicode dictionary keys are not allowed.
+        order_by_list = dictionary.get('order_by', [])
+        order_by_list = (_unicode_keys_to_strings(o) for o in order_by_list)
+        order_by = [OrderBy(**o) for o in order_by_list]
         limit = dictionary.get('limit')
         offset = dictionary.get('offset')
         return SearchParameters(filters=filters, limit=limit, offset=offset,
@@ -293,9 +309,8 @@ class QueryBuilder(object):
         field = getattr(model, relation or fieldname)
         return OPERATORS.get(operator)(field, argument, fieldname)
 
-    # TODO rename this to _create_filters
     @staticmethod
-    def _extract_operators(model, search_params):
+    def _create_filters(model, search_params):
         """Returns the list of operations on ``model`` specified in the
         :attr:`filters` attribute on the ``search_params`` object.
 
@@ -303,32 +318,21 @@ class QueryBuilder(object):
         class whose fields represent the parameters of the search.
 
         """
-        operations = []
-        for f in search_params.filters:
-            fname = f.fieldname
-            val = f.argument
-
+        filters = []
+        for filt in search_params.filters:
+            fname = filt.fieldname
+            val = filt.argument
             # get the relationship from the field name, if it exists
-            # TODO document that field names must not contain "__"
             relation = None
             if '__' in fname:
                 relation, fname = fname.split('__')
-
-            # for the sake of brevity
-            makeop = QueryBuilder._create_operation
-            # We are going to compare a field with another one, so there's
-            # no reason to parse
-            if f.otherfield:
-                otherfield = getattr(model, f.otherfield)
-                param = makeop(model, fname, f.operator, otherfield, relation)
-                operations.append(param)
-                continue
-
-            # Collecting the query
-            param = makeop(model, fname, f.operator, val, relation)
-            operations.append(param)
-
-        return operations
+            # get the other field to which to compare, if it exists
+            if filt.otherfield:
+                val = getattr(model, filt.otherfield)
+            param = QueryBuilder._create_operation(model, fname, filt.operator,
+                                                   val, relation)
+            filters.append(param)
+        return filters
 
     @staticmethod
     def create_query(model, search_params):
@@ -353,9 +357,9 @@ class QueryBuilder(object):
         """
         # Adding field filters
         query = model.query
-        operations = QueryBuilder._extract_operators(model, search_params)
-        for i in operations:
-            query = query.filter(i)
+        filters = QueryBuilder._create_filters(model, search_params)
+        for filt in filters:
+            query = query.filter(filt)
 
         # Order the search
         for val in search_params.order_by:
@@ -371,59 +375,6 @@ class QueryBuilder(object):
         return query
 
 
-def evaluate_functions(model, functions):
-    """Executes the each of the SQLAlchemy functions specified in
-    ``functions``, a list of dictionaries of the form described below, on the
-    given model and returns a dictionary mapping function name (slightly
-    modified, see below) to result of evaluation of that function.
-
-    ``functions`` is a list of dictionaries of the form::
-
-        {'name': 'avg', 'field': 'amount'}
-
-    For example, if you want the sum and the average of the field named
-    "amount"::
-
-        >>> # assume instances of Person exist in the database...
-        >>> f1 = dict(name='sum', field='amount')
-        >>> f2 = dict(name='avg', field='amount')
-        >>> evaluate_functions(Person, [f1, f2])
-        {'avg__amount': 456, 'sum__amount': 123}
-
-    The return value is a dictionary mapping ``'<funcname>__<fieldname>'`` to
-    the result of evaluating that function on that field. If `model` is
-    ``None`` or `functions` is empty, this function returns the empty
-    dictionary.
-
-    If a field does not exist on a given model, :exc:`AttributeError` is
-    raised. If a function does not exist,
-    :exc:`sqlalchemy.exc.OperationalError` is raised.
-
-    """
-    if not model or not functions:
-        return {}
-    processed = []
-    funcnames = []
-    for f in functions:
-        # We retrieve the function by name from the SQLAlchemy ``func``
-        # module and the field by name from the model class.
-        #
-        # If the specified function doesn't exist, this raises
-        # OperationalError. If the specified field doesn't exist, this raises
-        # AttributeError.
-        funcobj = getattr(func, f['name'])
-        field = getattr(model, f['field'])
-        # Time to store things to be executed. The processed list stores
-        # functions that will be executed in the database and funcnames
-        # contains names of the entries that will be returned to the
-        # caller.
-        funcnames.append('{}__{}'.format(f['name'], f['field']))
-        processed.append(funcobj(field))
-    # evaluate all the functions at once and get an iterable of results
-    evaluated = session.query(*processed).one()
-    return dict(zip(funcnames, evaluated))
-
-
 def create_query(model, searchparams):
     """Returns a SQLAlchemy query object on the given ``model`` where the
     search for the query is defined by ``searchparams``.
